/*!**************************************************************************
\file    AuxFuncs.h

*****************************************************************************/
#ifndef __AuxFuncs_H__
#define __AuxFuncs_H__

#include <cassert>
#include <cfloat>
#include <cmath>

//#################################################################################
//! \brief Global variable definitions
//#################################################################################

#define FLOATMAX ((float)(1e15))
#define MAXINT ((int)(32700))
#define EPSILON 3.0e-7

#ifdef _MSC_VER
#define log2(A) log10((float)A)/log10(2.0f)
//#define isfinite(A) (!(_isnan(A)))
#endif

//#define isfiniteBTF(A) (!(isnanf(A)))
#define isfiniteBTF(A) (isfinite(A))

inline float luminance(const float RGB[3])
{
  assert(RGB);
  // coefficients to compute luminance from RGB
  const float wlum[3] = {0.299f,  // R
			 0.587f,  // G
			 0.114f}; // B
  float lum = wlum[0] * RGB[0] + wlum[1] * RGB[1] + wlum[2] * RGB[2];
  return lum;  
}

inline double luminance(const double RGB[3])
{
  assert(RGB);
  // coefficients to compute luminance from RGB
  const double wlum[3] = {0.299,  // R
			 0.587,  // G
			 0.114}; // B
  double lum = wlum[0] * RGB[0] + wlum[1] * RGB[1] + wlum[2] * RGB[2];
  return lum;  
}

// from phi and theta to unit 3D vector. Z-axis points to the north pole
// of the sphere
void
ConvertThetaPhiToXYZ(float theta, float phi, float xyz[]);

// from phi and theta to unit 3D vector. Z-axis points to the north pole
// of the sphere
void
ConvertXYZtoThetaPhi(float xyz[], float &theta, float &phi);

// Computes arbitrary normal
void
ArbitraryNormal(const float N[3], float outN[3]);

// statistics record in C++, see Donald Knuth's book
struct SVAR
{
  float statsV; // computed statistics
  int cnt; // the number of samples taken to compute statistics
  double mean; // mean value1
  double M2; // sum for variance1
  // Statistical support
  SVAR() { reset();}

  // add a single sample
  void update(const double newSampleValue) {
    cnt++;
    double delta = newSampleValue - mean;
    mean += delta/(double)cnt;
    M2 += delta*(newSampleValue - mean);
    // temporarily for debugging
    evaluate();
  }
  // add a single zero
  void updateZero() {
    // we count also for zero values
    cnt++;
    double delta = - mean;
    mean += delta/(double)cnt;
    M2 -= delta*mean;
    // temporarily for debugging
    evaluate();
  }
  // Adding zero value
  void updateZeroXtimes(int cntZeros) {
    assert(cntZeros > 0);
    // naive solution - another faster to be developped
    // based on the variance of two random variables should be easy
    for (int i = 0; i < cntZeros; i++) {
      cnt++;
      double delta = - mean;
      mean += delta/(double)cnt;
      M2 += delta*(-mean);      
    }
    // temporarily for debugging
    evaluate();
  }
  void reset() { statsV = 0; cnt = 0; mean=M2=0; }
  // It returns unbised sample variance (so not for finite population)
  double evaluate() {
    return (statsV = (double)M2/((double)cnt - 1));
  }
  // It returns variance for finite population
  double evaluateFP() {
    return (statsV = (double)M2/((double)cnt));
  }
};

// The following functions below assume 3D vectors in array
// void vec_neg(float *x);
// void vec_diff(float *x, float *y, float *res);
// float vec_len(float *x);
// float vec_dist(float *x, float *y);
// int vec_norm(float *x);
// float vec_dotprod(float *x, float *y);
// void vec_crossprod(float *x, float *y, float *res);
// float point_dist(float *x, float *y);
// void normal(float *v1, float *v2, float *v3, float *n);
// float vec_angle(float *x, float *y);
// void vec_reflect(float *x, float *normal, float *res);
// float rad2deg(float d);
// float deg2rad(float r);

inline void vec_neg(float *x)
{
    x[0] = -x[0];
    x[1] = -x[1];
    x[2] = -x[2];
}//--- vec_neg ----------------------------------------------------

inline void vec_diff(float *x, float *y, float *res)
{
  res[0] = x[0] - y[0];
  res[1] = x[1] - y[1];
  res[2] = x[2] - y[2];
}//--- vec_diff ----------------------------------------------------

inline void vec_crossprod(float *x, float *y, float *res)
{
  res[0] = x[1]*y[2] - x[2]*y[1];
  res[1] = x[2]*y[0] - x[0]*y[2];
  res[2] = x[0]*y[1] - x[1]*y[0];
}//--- vec_crossprod ----------------------------------------------

inline float vec_len(float *x)
{
  return sqrtf(x[0]*x[0] + x[1]*x[1] + x[2]*x[2] + 1e-31);
}//--- vec_len ----------------------------------------------------

inline float vec_dist(float *x, float *y)
{
    return sqrtf((x[0]-y[0])*(x[0]-y[0]) + (x[1]-y[1])*(x[1]-y[1]) + (x[2]-y[2])*(x[2]-y[2]));
}//--- vec_dist ---------------------------------------------------

int vec_norm(float *x)
{
  float aux;
  aux = vec_len(x);
  if(aux == 0.f)
    {		
      x[0] = 0.f;
      x[1] = 0.f;
      x[2] = 0.f;
      printf("vec_norm: zero length vector");    
      return 0;
    }
  x[0] = x[0]/aux;
  x[1] = x[1]/aux;
  x[2] = x[2]/aux;
  return 1;
}//--- vec_norm ---------------------------------------------------

inline float vec_dotprod(float *x, float *y)
{
  return x[0]*y[0]+x[1]*y[1]+x[2]*y[2];
}//--- vec+dotprod ------------------------------------------------

inline float point_dist(float *x, float *y)
{
    float a,b,c;
    
    a = (x[0]-y[0])*(x[0]-y[0]);
    b = (x[1]-y[1])*(x[1]-y[1]);
    c = (x[2]-y[2])*(x[2]-y[2]);
    return sqrtf(a+b+c);
}//--- point_dist -------------------------------------------------

inline void normal(float *v1, float *v2, float *v3, float *n)
{
    float p[3], q[3];
    
    vec_diff(v1,v2,p);
    vec_diff(v3,v2,q);
    vec_crossprod(q,p,n);
    vec_norm(n);
}//--- normal -----------------------------------------------------

inline void face_center3(float *v1, float *v2, float *v3, float *fc)
{
    fc[0] = (v1[0]+v2[0]+v3[0])/3;
    fc[1] = (v1[1]+v2[1]+v3[1])/3;
    fc[2] = (v1[2]+v2[2]+v3[2])/3;    
}//--- face_center3 -----------------------------------------------

inline void face_center4(float *v1, float  *v2, float *v3, float *v4, float *fc)
{
    fc[0] = (v1[0]+v2[0]+v3[0]+v4[0])/4;
    fc[1] = (v1[1]+v2[1]+v3[1]+v4[1])/4;
    fc[2] = (v1[2]+v2[2]+v3[2]+v4[2])/4;    
}//--- face_center4 -----------------------------------------------

inline float vec_angle(float *x, float *y)
{
  float num,den, res;
  num = x[0]*y[0]+x[1]*y[1]+x[2]*y[2];
  den = vec_len(x)*vec_len(y);
  if(den==0)
    printf("vec_angle: Vector have zero length\n");
  if (den > 0) {
    res = num/den;
    if(res>1)
      res = 1.0;
    return acos(res);
  }
  return 0;
}//--- vec_angle --------------------------------------------------

//! computes ideally reflected ray for x given normal, putting the output to 'res'
inline void vec_reflect(float *x, float *normal, float *res)
{
  float xx[3],nn[3];
  xx[0] = x[0];
  xx[1] = x[1];
  xx[2] = x[2];
  nn[0] = normal[0];
  nn[1] = normal[1];
  nn[2] = normal[2];
  vec_norm(xx);
  vec_norm(nn);
  float dp = vec_dotprod(xx, nn);
  res[0] = xx[0] - 2.f*nn[0]*dp;
  res[1] = xx[1] - 2.f*nn[1]*dp;
  res[2] = xx[2] - 2.f*nn[2]*dp;
}

inline float rad2deg(float d)
{
    return (180.f/M_PI)*d;
}//--- rad2deg ----------------------------------------------------

inline float deg2rad(float r)
{
    return (M_PI/180.f)*r;
}//--- deg2rad ----------------------------------------------------

inline void angles2xyz(float theta, float phi, float *x, float *y, float *z)
{
  *x = sin(theta)*cos(phi);
  *y = sin(theta)*sin(phi);
  *z = cos(theta);
}//-- angles2num --------------------------------------------------


#endif // __AuxFuncs_H__
