// Written by Vlastimil Havran, 2013

#ifndef __BRDFMEASURED_H__
#define __BRDFMEASURED_H__

#include <cassert>
#include <cmath>
#include <iostream>
#include <fstream>

#include "configh.h"

__BEGIN_SBRDFLIB_HEADER

class BRDFmeasuredModel 
{
protected:
  int modefs; // 0 .. intensity, 1 .. spectrum
  int scanningAxis;
  // Wavelength for spectras
  int wavelengthCount; // .. the number of spectras measured
  float wavelengthBeg;
  float wavelengthStep;
  float wavelengthEnd;
  // Parameters in the file for theta of photodetector
  float thetaBeg;
  float thetaEnd;
  float thetaStep;
  int   thetaCount;
  int   thetaM_Count;
  // Parameters in the file for phi of photodetector
  float phiBeg;
  float phiEnd;
  float phiEndMax;
  float phiStep;
  int   phiCount;
  int   phiM_Count;
  // Parameters in the file for theta of illumination
  float thetaLightingBeg;
  float thetaLightingEnd;
  float thetaLightingStep;
  int   thetaLightingCount;
  int   thetaLightingM_Count;
  // the number of files to be read
  int numberOfFiles;

  // The parameters for the organization in the array
  int thetaMin, thetaMax; // theta of photodetector values
  int thetaMaxIndex; // theta of photodector index, max value
  int phiMin, phiMax; // phi of photodetector values
  int phiMaxIndex; // phi of photodector index, max value
  int thetaIlMin, thetaIlMax; // theta of illumination values
  int thetaIlMaxIndex; // theta of illumination index, max value

  // Read param from input file
  int ReadParam(std::fstream &fs, const char * keyword, int &parVar);
  int ReadParam(std::fstream &fs, const char * keyword, float &parVar);
  void ReadIntensity(std::fstream &fs, int tli);
  void ReadSpectrum(std::fstream &fs, int tli);

  float *vals; // values as read from the file
  float *rgbvals; // recomputed values to RGB, if spectral measurement
  int maxvindex, maxvrgbindex;
  int addr(int phiIndex, int thetaIndex, int thetaLightIndex) const {
    return thetaLightIndex*thetaM_Count*phiM_Count +
           thetaIndex*phiM_Count + phiIndex;
  }
  void setv(int phiIndex, int thetaIndex, int thetaLightIndex, float nval) {
    int vi=addr(phiIndex, thetaIndex, thetaLightIndex);
    assert(vi>=0); assert(vi<maxvindex);
    vals[vi] = nval;
  }
  float getv(int phiIndex, int thetaIndex, int thetaLightIndex) const {
    int vi=addr(phiIndex, thetaIndex, thetaLightIndex);
    assert(vi>=0); assert(vi<maxvindex);
    return vals[vi];
  }
  void getv(int phiIndex, int thetaIndex, int thetaLightIndex, float RGB[]) const {
    assert(rgbvals);
    int vi=3*addr(phiIndex, thetaIndex, thetaLightIndex);
    assert(vi>=0);
    assert(vi+2<maxvrgbindex);
    RGB[0] = rgbvals[vi]; 
    RGB[1] = rgbvals[vi+1]; 
    RGB[2] = rgbvals[vi+2]; 
    return;
  }
  // Get theta value from the index
  float GetValThetaLighting(int ti) const {
    assert(0 <= ti); assert(ti < thetaLightingCount);
    return thetaLightingBeg + thetaLightingStep*ti;
  }
  int GetIndexThetaLighting(float val) const {
    assert(thetaLightingBeg <= val); assert(val <= thetaLightingEnd);
    return (val - thetaLightingBeg)/thetaLightingStep;
  }
  // Get theta value from the index
  float GetValTheta(int ti) const {
    assert(0 <= ti); assert(ti < thetaCount);
    return thetaBeg + thetaStep*ti;
  }
  int GetIndexTheta(float val) const {
    assert(thetaBeg <= val); assert(val <= thetaEnd);
    return (val - thetaBeg)/thetaStep;
  }
  // Get theta value from the index
  float GetValPhi(int phi) const {
    assert(0 <= phi); assert(phi < phiCount);
    return phiBeg + phiStep*phi;
  }
  int GetIndexPhi(float val) const {
    assert(phiBeg <= val); assert(val <= phiEnd);
    return (val - phiBeg)/phiStep;
  }
  // Get theta value from the index
  float GetValWavelength(int wi) const {
    assert(0 <= wi); assert(wi < wavelengthCount);
    return wavelengthBeg + wavelengthStep*wi;
  }
  int GetIndexWavelength(float val) const {
    assert(wavelengthBeg <= val); assert(val <= wavelengthEnd);
    return (val - wavelengthBeg)/wavelengthStep;
  }
public:
 BRDFmeasuredModel():vals(0),rgbvals(0),maxvindex(-1) {}
  ~BRDFmeasuredModel();

  int ReadBRDFDataFromFile(char *filename);

  // This uses the measured data from the file
  void GetBRDF(float RGB[3],
	       float cosangle,
	       float refl_illum_vec[3],
	       float view_vec[3]) const;
  
  void GetBRDF(float RGB[3], // output RGB
	       float theta_illum, // illumination vector
	       float phi_illum,
	       float theta_view, // view vector
	       float phi_view) const;

  // query functions
  float ThetaLightingBeg() const { return thetaLightingBeg;}
  float ThetaLightingEnd() const { return thetaLightingEnd;}
  float ThetaLightingCount() const { return thetaLightingCount;}

  float ThetaBeg() const { return thetaBeg;}
  float ThetaEnd() const { return thetaEnd;}
  float ThetaCount() const { return thetaCount;}

  float PhiBeg() const { return phiBeg;}
  float PhiEnd() const { return phiEnd;}
  float PhiCount() const { return phiCount;}

  // Only for spectral measurement
  float WavelengthBeg() const {
    if (IsSpectralMeasurement())
      return wavelengthBeg;
    return 550; //[nm]
  }
  float WavelengthEnd() const {
    if (IsSpectralMeasurement())
      return wavelengthEnd;
    return 550; //[nm] 
  }
  float WavelengthCount() const {
    if (IsSpectralMeasurement())
      return wavelengthCount;
    return 1; // only 1 for intensity
  }
  float GetVal(int i) const { return vals[i];}

  bool IsSpectralMeasurement() const { return (modefs == 1) ? true : false; }

  // Convert to RGB
  void ConvertToRGB();
};

class BRDFmeasuredModelComposed:
  public BRDFmeasuredModel  
{
  bool active;
  // BRDFmeasuredModel brdf; // standard one is included in the ancestor function
  BRDFmeasuredModel *brdf2; // for the operation like minus  
public:
  BRDFmeasuredModelComposed():active(false),brdf2(0) {}
  ~BRDFmeasuredModelComposed();

  // Only the reading of two files is changed
  int ReadBRDFDataFromTwoFiles(char *filename1, char *filename2);
};

__END_SBRDFLIB_HEADER

#endif // __BRDFMEASURED_H__
