#ifndef _BRDFMODEL_H__
#define _BRDFMODEL_H__

#include <cstring>
#include <string.h>

#include "configh.h"

__BEGIN_SBRDFLIB_HEADER

// -------------------------------------------------------------------
// Common interface to BRDF used in this project
class BRDF
{
  mutable int cntThresholdMean; // minimum number of samples to get albedo/variance for a view_direction
  // statistics record for some theta_v,phi_v
  struct SEntryVA{
    float theta; // the theta_view used
    float phi; // the phi_view used
    float albedo; // computed albedo    
    float statsV; // computed statistics
    int cnt; // the number of samples taken to compute statistics
    double mean; // mean value1
    double M2; // sum for variance1
    double mean2; // mean value2
    double M22; // sum for variance2    
    float maxlum2; // the maximum value2
    SEntryVA() { reset();}
    // Statistical support
    void update(const double newValue, const double newValue2) {
      cnt++;
      // the first variable
      double delta = newValue - mean;
      mean += delta/(double)cnt;
      M2 += delta*(newValue - mean);
      // the second value 
      delta = newValue2 - mean2;
      mean2 += delta/(double)cnt;
      M22 += delta*(newValue2 - mean2);
      // update maximum for rejection sampling
      if (newValue2 > maxlum2)
	maxlum2 = newValue2; 
    }
    void reset() { theta=-1; phi=0; albedo=statsV=0; cnt = 0; mean=M2=mean2=M22=maxlum2=0; }
    void evaluate() {
      albedo = M_PI * 2.0f * mean;
      if (albedo > 1.0) albedo = 1.0f;
      statsV = (double)M22/((double)cnt - 1);
    }
  };
  // individual sampling
  mutable SEntryVA individualSVA; // for individual query without interpolation
  enum { // minimum number of samples to get albedo/variance for a view_direction
    cntThresholdMeanInit = 100
  }; 
  // Get albedo for viewing direction
  virtual float getAlbedoSampling(int irow, int jcol, // texture position
				  // viewer direction as two angles in radians
				  float theta_view, float phi_view) const;
  // compute albedo in simple way
  void getAlbedoSamplingSimple(int irow, int jcol, // texture position
			       // viewer direction as two angles in radians
			       float theta_view, float phi_view,
			       SEntryVA &sva) const;
public:
  BRDF() {}
  virtual ~BRDF() {}
  // the deep copy operator
  virtual BRDF *Dup() const = 0;

  // returns cosine corrected value of BRDF in RGB (corresponding to non-normalized PDF) given
  // the illumination and viewer direction, returns false on success
  virtual bool getBRDFValue(int irow, int jcol, // texture position
			    float theta_view, float phi_view, // viewer direction as two angles in radians
			    float theta_il, float phi_il, // illumination direction as two angles in radians
			    float RGB[]) const = 0; // output value
  
  // Get albedo for viewing direction
  virtual float getAlbedo(int irow, int jcol, // texture position
			  // viewer direction as two angles in radians
			  float theta_view, float phi_view) const;

  // Make stochastic sampling, where Z-axis is the normal of a surface
  virtual bool spawnStochasticRay(int irow, int jcol,
				  float rnd1, float rnd2, // input
				  float thetaV, float phiV, // the direction of a viewer
				  float XYZ[3], // output direction
				  float RGB[], // corresponding color
				  float &prob) const;

  // As default BRDF functions are considered isotropic .. returns false here
  virtual bool isAnisotropic() const {return false;}

  // As default BRDF functions are considered non spatially varying
  virtual bool isSpatiallyVarying() const {return false;}

  enum BrdfErrors {
    EE_NoError = 0,
    EE_OutOfRange = 1,
  };

// this enum is used to determine type of the parameter, when the
  // method GetParameterSyntax() is called.  
  enum EParameterUIType
  {
	// unknown type of the parameter
	EE_Unknown  = 0x0000,
	// one float - x
	EE_Float	= 0x0001,
	// two floats - x,y
	EE_XY		= 0x0002,
	// B component of the color
	EE_ColorB   = 0x0004,
	// RGB color
	EE_ColorRGB = 0x0008,
	// RGB components parameter
	EE_RGB_Components = 0x0010,
	// boolean value
	EE_Boolean  = 0x0020,
	// 3-component values (x,y,z)
	EE_XYZ = 0x0040,
	// 4-component values (x,y,z and weight)
	EE_Weighted_XYZ = 0x0080,
	// 4-component values (x,y,z,w)
	EE_XYZW = 0x0100,
	// Special type of texture consisting of several basic types
	EE_Fresnel = 0x0200
  };
  
  // Different BRDF models
  enum {
    EE_Phong = 1,
    EE_Matusik = 2,
  };

  // returns probability of sampling in a given viewer direction
  // The user has to provide also RGB[] since it is computed from
  // the value. Typically in MIS, this function is called after
  // the call getBRDFValue() that RGB is computed. RGB is not changed.
  virtual float getProbOfSampling(int irow, int jcol, // texture position
				  float theta_view, float phiV, // viewer direction as two angles in radians
				  float theta_i, float phi_i, // illumination direction as two angles in radians
				  const float RGB[]) const; // corresponding color
  
  virtual int getID() const = 0; // Get ID of the BRDF class

   /**
     Returns a string key to detailed description of the current BRDF model. Its an
	 unique id in a string form, which identifies a BRDF model. Based on this string key,
	 detailed description containing meaning of the model, its purpose, typical 
	 usability, etc. is displayed to the user and so it helps to better understand the 
	 purpose of the model.

     @return				string with the description of the model
  */
  const char* getHelp() const;
  
  // Provides the description of BRDF model to a string
  virtual void getString(string &outstr) const = 0; // abstract method

  /**
     Gets the number of parameters of the current model

	 @return				parameters count
  */
  virtual int 
  getNumParameters() const = 0;

  // manipulation with BRDF model parameters - connection with possible GUI

  /**
     Returns the title of the parameter based on index of this parameter
     @param parameterIndex	index of the parameter whose title we want to know
     @return				string with the title of the parameter
  */
  virtual const char* 
  getParameterTitle(int parameterIndex) const;

  /**
     Returns a description of the parameter based on index of this parameter.
	 It describes meaning of the target parameter in detail and its purpose and practice 
	 influence to appearance of the visualisation of the model.
     @param parameterIndex	index of the parameter whose description we want to know
     @return				string with the description of the parameter
  */
  virtual const char* 
  getParameterDescription(int parameterIndex) const = 0;

  /**
     Returns syntax of the parameter based on index of this parameter.
	 Syntax means type of the parameter and it helps programmer to dynamically
	 create GUI element for this parameter based on its type (e.g. R/G/B values
	 have different GUI inputs than some float parameter). So it helps programmer
	 to distinguish different types of parameters between each other. See
	 EParameterUIType for these parameter types.

	 @param parameterIndex	index of the parameter which syntax we want to know
	 @return		string with the description of the model
  */
  virtual EParameterUIType
  getParameterSyntax(int parameterIndex) const = 0;

   /**
     Gets current value of the target parameter

     @param parameterIndex(in)	index of the parameter which value we want to know

     @return			the number of parameters in the values set
  */
  virtual int
  getParameter(int parameterIndex, float values[]) const = 0;

  /**
     Sets value of the target parameter

     @param parameterIndex(in)	index of the parameter which value we want to set
     @param value(in)			value to set to the parameter
     @return                    the number of values set
  */
  virtual int
  setParameter(int parameterIndex, float newvalue[]) = 0;

  /// Gets bounds of the parameter and its default value
  /**
     Sets minimum and maximum bound of the parameter and its initial value and
     returns these values in the vector in this order : min,max,default

     @param parameterIndex(in)	index of the parameter whose contraints we want to get

     @return			vector of contraints of the target parameter
  */
  virtual int
  getParameterConstraints(int parameterIndex, float minvals[], float maxvals[]) const = 0;
};

extern int errorSBRDFLIB;

__BEGIN_SBRDFLIB_HEADER

#endif // _BRDFMODEL_H__
